/*
 * This file is part of Skyward Hub.
 *
 * Skyward Hub is free software: you can redistribute it and/or modify it under
 * the terms of the GNU General Public License as published by the Free Software
 * Foundation, either version 3 of the License, or (at your option) any later
 * version.
 *
 * Skyward Hub is distributed in the hope that it will be useful, but WITHOUT
 * ANY WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS
 * FOR A PARTICULAR PURPOSE. See the GNU General Public License for more
 * details.
 *
 * You should have received a copy of the GNU General Public License along with
 * Skyward Hub. If not, see <https://www.gnu.org/licenses/>.
 *
 */

#include "ARPStateViewer.h"

#include <Components/FilterSelector/FilterSelector.h>
#include <Core/MessageBroker/MessageBroker.h>

ARPStateViewerModule::ARPStateViewerModule()
    : Module(ModuleId::ARP_STATE_VIEWER)
{
    setupUi();
    customContextMenuActionSetup();
}

ARPStateViewerModule::~ARPStateViewerModule()
{
    MessageBroker::getInstance().unsubscribe(filter, this);
}

XmlObject ARPStateViewerModule::toXmlObject()
{
    XmlObject obj = Module::toXmlObject();

    obj.addAttribute("filter", filter.toString());

    return obj;
}

void ARPStateViewerModule::fromXmlObject(const XmlObject& xmlObject)
{
    auto filter = Filter::fromString(xmlObject.getAttribute("filter"));
    setFilter(filter);
}

void ARPStateViewerModule::setupUi()
{
    outerLayout = new QHBoxLayout;
    outerLayout->setContentsMargins(0, 0, 0, 0);
    outerLayout->setSpacing(0);

    for (auto labelText : ARPStatesList::statesLabels)
    {
        QLabel* label = new QLabel;
        label->setText(labelText);
        label->setAlignment(Qt::AlignCenter);
        label->setContentsMargins(4, 4, 4, 4);
        outerLayout->addWidget(label);
    }

    setLayout(outerLayout);
}

void ARPStateViewerModule::customContextMenuActionSetup()
{
    QAction* action = new QAction("Choose topic and field");
    connect(action, &QAction::triggered, this,
            &ARPStateViewerModule::onConfigureClicked);

    customContextMenuActions.append(action);
}

void ARPStateViewerModule::onConfigureClicked()
{
    FilterSelector::selectFilter(
        filter, [this](const Filter& newFilter) { setFilter(newFilter); });
}

void ARPStateViewerModule::setFilter(const Filter& newFilter)
{
    MessageBroker::getInstance().unsubscribe(filter, this);
    MessageBroker::getInstance().subscribe(
        newFilter, this,
        [this](const Message& message, const Filter& filter)
        { onMsgReceived(message); });
    filter = newFilter;
}

void ARPStateViewerModule::onMsgReceived(const Message& msg)
{
    Field field = msg.getField("state");

    ARPStatesList::State state =
        static_cast<ARPStatesList::State>(field.getUnsignedInteger());

    // Skip invalid and macro states
    if (state == ARPStatesList::State::INVALID ||
        state == ARPStatesList::State::CONFIG ||
        state == ARPStatesList::State::FEEDBACK ||
        state == ARPStatesList::State::NO_FEEDBACK)
        return;

    if (state != currentState)
    {
        currentState = static_cast<ARPStatesList::State>(state);

        QString baseStyle =
            "border-bottom-width:1px;border-left-width:1px;border-top-width:"
            "1px;border-radius:0;";
        QString currentStateStyle =
            "background-color:yellow; color:black;" + baseStyle;
        QString completedStyle = "background-color:green;" + baseStyle;
        QString errorStyle     = "background-color:red;" + baseStyle;
        QString nfStyle =
            "background-color:yellow;"
            "border-bottom-width:4px;border-left-width:4px;"
            "border-right-width:4px; border-top-width:4px;"
            "border-radius:0; border-style: dashed;"
            "border-color:red;";

        int index  = ARPStatesList::statesIndexes[currentState];
        auto label = outerLayout->itemAt(index)->widget();

        if (currentState == ARPStatesList::State::INIT_ERROR)
            label->setStyleSheet(errorStyle);
        else if (currentState == ARPStatesList::State::ACTIVE_NF ||
                 currentState == ARPStatesList::State::ARM_READY ||
                 currentState == ARPStatesList::State::ARMED_NF ||
                 currentState == ARPStatesList::State::FIX_ROCKET_NF ||
                 currentState == ARPStatesList::State::INSERT_INFO ||
                 currentState == ARPStatesList::State::TEST_NF)
            label->setStyleSheet(nfStyle);
        else
            label->setStyleSheet(currentStateStyle);

        // Set as completed every state before the current
        for (int i = index - 1; i >= 0; i--)
            outerLayout->itemAt(i)->widget()->setStyleSheet(completedStyle);

        // Reset every state after the current
        for (int i = index + 1; i < ARPStatesList::statesLabels.count(); i++)
            outerLayout->itemAt(i)->widget()->setStyleSheet(baseStyle);

        auto tmp = outerLayout->itemAt(ARPStatesList::statesLabels.count() - 1)
                       ->widget()
                       ->styleSheet();
        outerLayout->itemAt(ARPStatesList::statesLabels.count() - 1)
            ->widget()
            ->setStyleSheet(tmp + "border-right-width:1px;");
    }
}
