/*
 * This file is part of Skyward Hub.
 *
 * Skyward Hub is free software: you can redistribute it and/or modify it under
 * the terms of the GNU General Public License as published by the Free Software
 * Foundation, either version 3 of the License, or (at your option) any later
 * version.
 *
 * Skyward Hub is distributed in the hope that it will be useful, but WITHOUT
 * ANY WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS
 * FOR A PARTICULAR PURPOSE. See the GNU General Public License for more
 * details.
 *
 * You should have received a copy of the GNU General Public License along with
 * Skyward Hub. If not, see <https://www.gnu.org/licenses/>.
 *
 */

#include "MessageFormElement.h"

#include <Modules/SkywardHubStrings.h>

MessageFormElement::MessageFormElement() {}

MessageFormElement::~MessageFormElement()
{
    for (auto key : comboBoxMap.keys())
    {
        delete comboBoxMap[key].first;
        delete comboBoxMap[key].second;
    }

    for (auto key : floatMap.keys())
    {
        delete std::get<0>(floatMap[key]);
        delete std::get<1>(floatMap[key]);
    }

    for (auto key : integerMap.keys())
    {
        delete std::get<0>(integerMap[key]);
        delete std::get<1>(integerMap[key]);
    }
}

bool MessageFormElement::addComboBox(QString id, QString label,
                                     const QMap<QString, int>& options)
{
    if (!comboBoxMap.contains(id) && !floatMap.contains(id) &&
        !integerMap.contains(id))
    {
        auto* comboBox = new QComboBox;
        comboBox->setSizePolicy(
            QSizePolicy(QSizePolicy::Expanding, QSizePolicy::Minimum));

        for (auto key : options.keys())
            comboBox->addItem(key, options[key]);

        comboBoxMap[id] = {new QLabel(label), comboBox};
        return true;
    }
    else
    {
        return false;
    }
}

bool MessageFormElement::addFloat(QString id, QString label, float min,
                                  float max, int decimals)
{
    if (!comboBoxMap.contains(id) && !floatMap.contains(id) &&
        !integerMap.contains(id))
    {
        auto* lineEdit = new QLineEdit;
        lineEdit->setSizePolicy(
            QSizePolicy(QSizePolicy::Expanding, QSizePolicy::Minimum));
        lineEdit->setValidator(new QDoubleValidator(min, max, decimals));
        floatMap[id] = {new QLabel(label), lineEdit, min, max, decimals};
        return true;
    }
    else
    {
        return false;
    }
}

bool MessageFormElement::addInteger(QString id, QString label, int min, int max)
{
    if (!comboBoxMap.contains(id) && !floatMap.contains(id) &&
        !integerMap.contains(id))
    {
        auto* lineEdit = new QLineEdit;
        lineEdit->setSizePolicy(
            QSizePolicy(QSizePolicy::Expanding, QSizePolicy::Minimum));
        lineEdit->setValidator(new QIntValidator(min, max));
        integerMap[id] = {new QLabel(label), lineEdit, min, max};
        return true;
    }
    else
    {
        return false;
    }
}

XmlObject MessageFormElement::toXmlObject(XmlObject& obj)
{
    for (auto key : comboBoxMap.keys())
    {
        auto comboBox = comboBoxMap[key];
        XmlObject comboBoxXml("combo_box");
        comboBoxXml.addAttribute("id", key);
        comboBoxXml.addAttribute("text", comboBox.second->currentText());
        obj.addChild(comboBoxXml);
    }

    for (auto key : floatMap.keys())
    {
        auto floatField = floatMap[key];
        XmlObject floatFieldXml("float_field");
        floatFieldXml.addAttribute("id", key);
        floatFieldXml.addAttribute("text", std::get<1>(floatField)->text());
        floatFieldXml.addAttribute("min", std::get<2>(floatField));
        floatFieldXml.addAttribute("max", std::get<3>(floatField));
        floatFieldXml.addAttribute("decimals", std::get<4>(floatField));
        obj.addChild(floatFieldXml);
    }

    for (auto key : floatMap.keys())
    {
        auto integerField = floatMap[key];
        XmlObject integerFieldXml("integer_field");
        integerFieldXml.addAttribute("id", key);
        integerFieldXml.addAttribute("text", std::get<1>(integerField)->text());
        integerFieldXml.addAttribute("min", std::get<2>(integerField));
        integerFieldXml.addAttribute("max", std::get<3>(integerField));
        obj.addChild(integerFieldXml);
    }

    return obj;
}

void MessageFormElement::fromXmlObject(const XmlObject& obj)
{
    for (int i = 0; i < obj.childCount(); i++)
    {
        auto child = obj.childAt(i);
        auto key   = child.getAttribute("id");

        if (child.getObjectName() == "combo_box" && comboBoxMap.contains(key))
        {
            auto comboBox = comboBoxMap[key];
            comboBox.second->setCurrentText(child.getAttribute("text"));
        }
        else if (child.getObjectName() == "float_field" &&
                 floatMap.contains(key))
        {
            auto floatField = floatMap[key];
            std::get<1>(floatField)->setText(child.getAttribute("text"));
            child.getAttribute("min", std::get<2>(floatField));
            child.getAttribute("max", std::get<3>(floatField));
            child.getAttribute("decimals", std::get<4>(floatField));
            std::get<1>(floatField)
                ->setValidator(new QDoubleValidator(std::get<2>(floatField),
                                                    std::get<3>(floatField),
                                                    std::get<4>(floatField)));
        }
        else if (child.getObjectName() == "integer_field" &&
                 integerMap.contains(key))
        {
            auto integerField = integerMap[key];
            std::get<1>(integerField)->setText(child.getAttribute("text"));
            child.getAttribute("min", std::get<2>(integerField));
            child.getAttribute("max", std::get<3>(integerField));
            std::get<1>(integerField)
                ->setValidator(new QIntValidator(std::get<2>(integerField),
                                                 std::get<3>(integerField)));
        }
    }
}

void MessageFormElement::applyToGridLayout(QGridLayout* layout)
{
    for (auto key : comboBoxMap.keys())
    {
        auto label    = comboBoxMap[key].first;
        auto comboBox = comboBoxMap[key].second;

        int rows = layout->rowCount();
        layout->addWidget(label, rows, 0, Qt::AlignRight);
        layout->addWidget(comboBox, rows, 1);

        label->show();
        comboBox->show();
    }

    for (auto key : floatMap.keys())
    {
        auto label      = std::get<0>(floatMap[key]);
        auto floatField = std::get<1>(floatMap[key]);

        int rows = layout->rowCount();
        layout->addWidget(label, rows, 0, Qt::AlignRight);
        layout->addWidget(floatField, rows, 1);

        label->show();
        floatField->show();
    }

    for (auto key : integerMap.keys())
    {
        auto label        = std::get<0>(integerMap[key]);
        auto integerField = std::get<1>(integerMap[key]);

        int rows = layout->rowCount();
        layout->addWidget(label, rows, 0, Qt::AlignRight);
        layout->addWidget(integerField, rows, 1);

        label->show();
        integerField->show();
    }
}

void MessageFormElement::removeFromGridLayout(QGridLayout* layout)
{
    for (auto key : comboBoxMap.keys())
    {
        auto label    = comboBoxMap[key].first;
        auto comboBox = comboBoxMap[key].second;

        layout->removeWidget(label);
        layout->removeWidget(comboBox);

        label->hide();
        comboBox->hide();
    }

    for (auto key : floatMap.keys())
    {
        auto label      = std::get<0>(floatMap[key]);
        auto floatField = std::get<1>(floatMap[key]);

        layout->removeWidget(label);
        layout->removeWidget(floatField);

        label->hide();
        floatField->hide();
    }

    for (auto key : integerMap.keys())
    {
        auto label        = std::get<0>(integerMap[key]);
        auto integerField = std::get<1>(integerMap[key]);

        layout->removeWidget(label);
        layout->removeWidget(integerField);

        label->hide();
        integerField->hide();
    }
}

Message MessageFormElement::prepareMessage(const QString& sysName,
                                           const QString& messageId)
{
    Message message(Topic{sysName + "/Out/" + messageId});

    for (auto key : comboBoxMap.keys())
    {
        auto comboBox = comboBoxMap[key].second;
        auto msg      = Field{(uint64_t)comboBox->currentData().toInt()};
        message.setField(key, msg);
    }

    for (auto key : floatMap.keys())
    {
        auto floatField = std::get<1>(floatMap[key]);
        auto msg        = Field{floatField->text().toDouble()};
        message.setField(key, msg);
    }

    for (auto key : integerMap.keys())
    {
        auto integerField = std::get<1>(integerMap[key]);
        auto msg          = Field{(uint64_t)integerField->text().toInt()};
        message.setField(key, msg);
    }

    return message;
}
