/*
 * This file is part of Skyward Hub.
 *
 * Skyward Hub is free software: you can redistribute it and/or modify it under
 * the terms of the GNU General Public License as published by the Free Software
 * Foundation, either version 3 of the License, or (at your option) any later
 * version.
 *
 * Skyward Hub is distributed in the hope that it will be useful, but WITHOUT
 * ANY WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS
 * FOR A PARTICULAR PURPOSE. See the GNU General Public License for more
 * details.
 *
 * You should have received a copy of the GNU General Public License along with
 * Skyward Hub. If not, see <https://www.gnu.org/licenses/>.
 *
 */

#include "MavlinkCodec.h"

#include <qregularexpression.h>

#include <QDir>

#include "BaseMavlinkModule.h"
#include "Modules/SkywardHubStrings.h"

MavlinkCodec::MavlinkCodec(MavlinkPort* port, BaseMavlinkModule* parent)
    : port(port), parent(parent)
{
}

void MavlinkCodec::send(const mavlink_message_t& msg)
{
    unsigned char
        buf[MAVLINK_NUM_NON_PAYLOAD_BYTES + MAVLINK_MAX_DIALECT_PAYLOAD_SIZE];
    uint16_t len = mavlink_msg_to_send_buffer(buf, &msg);

    port->writeBytes(reinterpret_cast<char*>(buf), len);
}

void MavlinkCodec::startReading()
{
    QObject::connect(port, &MavlinkPort::bytesReceived, this,
                     &MavlinkCodec::onBytesReceived);
}

void MavlinkCodec::stopReading()
{
    QObject::disconnect(port, &MavlinkPort::bytesReceived, this,
                        &MavlinkCodec::onBytesReceived);
}

void MavlinkCodec::closeLog()
{
    if (logFile.isOpen())
        logFile.close();
}

void MavlinkCodec::openLogFile()
{
    closeLog();

    setLogFilePath();

    logFile.setFileName(logFilePath);
    if (logFile.open(QIODevice::WriteOnly))
    {
        if (parent != nullptr)
        {
            parent->info(tr("Mavlink"), tr("Opened file ") + logFilePath +
                                            tr(" for logging."));
        }
        QTextStream out(&logFile);
        out << "Log started " << QDateTime::currentDateTime().toString()
            << "\n";
    }
    else
    {
        if (parent != nullptr)
        {
            parent->error(tr("Mavlink"),
                          tr("Error, cannot open file ") + logFilePath);
        }
    }
}

const mavlink_message_info_t& MavlinkCodec::getMessageFormat(uint8_t messageId)
{
    static const mavlink_message_info_t infos[256] = MAVLINK_MESSAGE_INFO;
    return infos[messageId];
}

void MavlinkCodec::onBytesReceived(const char* data, qsizetype len)
{
    for (qsizetype i = 0; i < len; i++)
    {
        if (mavlink_parse_char(MAVLINK_COMM_0, data[i], &mavlinkMessage,
                               &mavlinkStatus))
            emit msgReceived(decodeMessage(mavlinkMessage));
    }

    if (logFile.isOpen())
        logFile.write(data, len);
}

void MavlinkCodec::setLogFilePath()
{
    QDir dir(SkywardHubStrings::defaultLogsFolder);
    if (!dir.exists())
        dir.mkpath(".");

    QString extension = ".dat";
    QString currentFile =
        QDateTime::currentDateTime().toString("yyyyMMdd_hh.mm.ss");
    QString proposedFilePath =
        SkywardHubStrings::defaultLogsFolder + "/" + currentFile;
    QString fileNumber = "";
    int i              = 1;
    while (QFile::exists(proposedFilePath + fileNumber + extension) && i < 200)
    {
        i++;
        fileNumber = " (" + QString::number(i) + ")";
    }

    logFilePath = proposedFilePath + fileNumber + extension;
}

Message MavlinkCodec::decodeMessage(const mavlink_message_t& msg)
{
    const mavlink_message_info_t& info = getMessageFormat(msg.msgid);

    QMap<QString, Field> fields;
    for (unsigned i = 0; i < info.num_fields; i++)
        fields[QString(info.fields[i].name)] = decodeField(msg, info.fields[i]);

    // Also embed system/component informations
    fields[SkywardHubStrings::mavlink_system_id_name] =
        Field(static_cast<uint64_t>(msg.sysid));
    fields[SkywardHubStrings::mavlink_component_id_name] =
        Field(static_cast<uint64_t>(msg.compid));

    Message output;
    output.setTopic(Topic(info.name));
    output.setFields(std::move(fields));
    return output;
}

Field MavlinkCodec::decodeField(const mavlink_message_t& msg,
                                const mavlink_field_info_t& field)
{
    if (field.array_length == 0)
    {
        return decodeArrayElement(msg, field, 0);
    }
    else
    {
        if (field.type == MAVLINK_TYPE_CHAR)
        {
            QString str;

            for (unsigned i = 0; i < field.array_length; i++)
            {
                str.append(_MAV_RETURN_char(&msg, field.wire_offset + i));

                if (_MAV_RETURN_char(&msg, field.wire_offset + i) == '\0')
                    break;
            }

            return Field(str);
        }
        else
        {
            return Field();
        }
    }
}

Field MavlinkCodec::decodeArrayElement(const mavlink_message_t& msg,
                                       const mavlink_field_info_t& field,
                                       int idx)
{
    switch (field.type)
    {
        case MAVLINK_TYPE_CHAR:
        {
            return Field(static_cast<uint64_t>(
                _MAV_RETURN_char(&msg, field.wire_offset + idx * 1)));
        }
        case MAVLINK_TYPE_UINT8_T:
        {
            return Field(static_cast<uint64_t>(
                _MAV_RETURN_uint8_t(&msg, field.wire_offset + idx * 1)));
        }
        case MAVLINK_TYPE_INT8_T:
        {
            return Field(static_cast<int64_t>(
                _MAV_RETURN_int8_t(&msg, field.wire_offset + idx * 1)));
        }
        case MAVLINK_TYPE_UINT16_T:
        {
            return Field(static_cast<uint64_t>(
                _MAV_RETURN_uint16_t(&msg, field.wire_offset + idx * 2)));
        }
        case MAVLINK_TYPE_INT16_T:
        {
            return Field(static_cast<int64_t>(
                _MAV_RETURN_int16_t(&msg, field.wire_offset + idx * 2)));
        }
        case MAVLINK_TYPE_UINT32_T:
        {
            return Field(static_cast<uint64_t>(
                _MAV_RETURN_uint32_t(&msg, field.wire_offset + idx * 4)));
        }
        case MAVLINK_TYPE_INT32_T:
        {
            return Field(static_cast<int64_t>(
                _MAV_RETURN_int32_t(&msg, field.wire_offset + idx * 4)));
        }
        case MAVLINK_TYPE_UINT64_T:
        {
            return Field(static_cast<uint64_t>(
                _MAV_RETURN_uint64_t(&msg, field.wire_offset + idx * 8)));
        }
        case MAVLINK_TYPE_INT64_T:
        {
            return Field(static_cast<int64_t>(
                _MAV_RETURN_int64_t(&msg, field.wire_offset + idx * 8)));
        }
        case MAVLINK_TYPE_FLOAT:
        {
            return Field(static_cast<double>(
                _MAV_RETURN_float(&msg, field.wire_offset + idx * 4)));
        }
        case MAVLINK_TYPE_DOUBLE:
        {
            return Field(static_cast<double>(
                _MAV_RETURN_double(&msg, field.wire_offset + idx * 8)));
        }
        default:
        {
            // Unsupported: return EMPTY
            return Field();
        }
    }
}

bool MavlinkCodec::encodeMessage(const Message& msg, SysId sysId, CompId compId,
                                 mavlink_message_t& output)
{
    QRegularExpression topicPattern("\\/((?:\\w|_)+)$");

    auto matches = topicPattern.match(msg.getTopic().toString());

    if (!matches.hasMatch())
    {
        return false;
    }

    QString messageName = matches.captured(1);

    if (messageName == "PING_TC")
    {
        mavlink_msg_ping_tc_pack(
            sysId, compId, &output,
            msg.getField("timestamp").getUnsignedInteger());
        return true;
    }
    else if (messageName == "COMMAND_TC")
    {
        mavlink_msg_command_tc_pack(
            sysId, compId, &output,
            msg.getField("command_id").getUnsignedInteger());
        return true;
    }
    else if (messageName == "ARP_COMMAND_TC")
    {
        mavlink_msg_arp_command_tc_pack(
            sysId, compId, &output,
            msg.getField("command_id").getUnsignedInteger());
        return true;
    }
    else if (messageName == "SYSTEM_TM_REQUEST_TC")
    {
        mavlink_msg_system_tm_request_tc_pack(
            sysId, compId, &output, msg.getField("tm_id").getUnsignedInteger());
        return true;
    }
    else if (messageName == "SENSOR_TM_REQUEST_TC")
    {
        mavlink_msg_sensor_tm_request_tc_pack(
            sysId, compId, &output,
            msg.getField("sensor_id").getUnsignedInteger());
        return true;
    }
    else if (messageName == "SERVO_TM_REQUEST_TC")
    {
        mavlink_msg_servo_tm_request_tc_pack(
            sysId, compId, &output,
            msg.getField("servo_id").getUnsignedInteger());
        return true;
    }
    else if (messageName == "SET_SERVO_ANGLE_TC")
    {
        mavlink_msg_set_servo_angle_tc_pack(
            sysId, compId, &output,
            msg.getField("servo_id").getUnsignedInteger(),
            msg.getField("angle").getDouble());
        return true;
    }
    else if (messageName == "WIGGLE_SERVO_TC")
    {
        mavlink_msg_wiggle_servo_tc_pack(
            sysId, compId, &output,
            msg.getField("servo_id").getUnsignedInteger());
        return true;
    }
    else if (messageName == "RESET_SERVO_TC")
    {
        mavlink_msg_reset_servo_tc_pack(
            sysId, compId, &output,
            msg.getField("servo_id").getUnsignedInteger());
        return true;
    }
    else if (messageName == "SET_REFERENCE_ALTITUDE_TC")
    {
        mavlink_msg_set_reference_altitude_tc_pack(
            sysId, compId, &output, msg.getField("ref_altitude").getDouble());
        return true;
    }
    else if (messageName == "SET_REFERENCE_TEMPERATURE_TC")
    {
        mavlink_msg_set_reference_temperature_tc_pack(
            sysId, compId, &output, msg.getField("ref_temp").getDouble());
        return true;
    }
    else if (messageName == "SET_ORIENTATION_TC")
    {
        mavlink_msg_set_orientation_tc_pack(sysId, compId, &output,
                                            msg.getField("yaw").getDouble(),
                                            msg.getField("pitch").getDouble(),
                                            msg.getField("roll").getDouble());
        return true;
    }
    else if (messageName == "SET_ORIENTATION_QUAT_TC")
    {
        mavlink_msg_set_orientation_quat_tc_pack(
            sysId, compId, &output, msg.getField("quat_x").getDouble(),
            msg.getField("quat_y").getDouble(),
            msg.getField("quat_z").getDouble(),
            msg.getField("quat_w").getDouble());
        return true;
    }
    else if (messageName == "SET_COORDINATES_TC")
    {
        mavlink_msg_set_coordinates_tc_pack(
            sysId, compId, &output, msg.getField("latitude").getDouble(),
            msg.getField("longitude").getDouble());
        return true;
    }
    else if (messageName == "RAW_EVENT_TC")
    {
        mavlink_msg_raw_event_tc_pack(
            sysId, compId, &output,
            msg.getField("topic_id").getUnsignedInteger(),
            msg.getField("event_id").getUnsignedInteger());
        return true;
    }
    else if (messageName == "SET_DEPLOYMENT_ALTITUDE_TC")
    {
        mavlink_msg_set_deployment_altitude_tc_pack(
            sysId, compId, &output, msg.getField("dpl_altitude").getDouble());
        return true;
    }
    else if (messageName == "SET_TARGET_COORDINATES_TC")
    {
        mavlink_msg_set_target_coordinates_tc_pack(
            sysId, compId, &output, msg.getField("latitude").getDouble(),
            msg.getField("longitude").getDouble());
        return true;
    }
    else if (messageName == "SET_ALGORITHM_TC")
    {
        mavlink_msg_set_algorithm_tc_pack(
            sysId, compId, &output,
            msg.getField("algorithm_number").getUnsignedInteger());
        return true;
    }
    else if (messageName == "SET_ATOMIC_VALVE_TIMING_TC")
    {
        mavlink_msg_set_atomic_valve_timing_tc_pack(
            sysId, compId, &output,
            msg.getField("servo_id").getUnsignedInteger(),
            msg.getField("maximum_timing").getUnsignedInteger());
        return true;
    }
    else if (messageName == "SET_VALVE_MAXIMUM_APERTURE_TC")
    {
        mavlink_msg_set_valve_maximum_aperture_tc_pack(
            sysId, compId, &output,
            msg.getField("servo_id").getUnsignedInteger(),
            msg.getField("maximum_aperture").getDouble());
        return true;
    }
    else if (messageName == "SET_IGNITION_TIME_TC")
    {
        mavlink_msg_set_ignition_time_tc_pack(
            sysId, compId, &output,
            msg.getField("timing").getUnsignedInteger());
        return true;
    }
    else if (messageName == "SET_NITROGEN_TIME_TC")
    {
        mavlink_msg_set_nitrogen_time_tc_pack(
            sysId, compId, &output,
            msg.getField("timing").getUnsignedInteger());
        return true;
    }
    else if (messageName == "SET_COOLING_TIME_TC")
    {
        mavlink_msg_set_cooling_time_tc_pack(
            sysId, compId, &output,
            msg.getField("timing").getUnsignedInteger());
        return true;
    }
    else if (messageName == "CONRIG_STATE_TC")
    {
        mavlink_msg_conrig_state_tc_pack(
            sysId, compId, &output,
            msg.getField("ignition_btn").getUnsignedInteger(),
            msg.getField("filling_valve_btn").getUnsignedInteger(),
            msg.getField("venting_valve_btn").getUnsignedInteger(),
            msg.getField("release_pressure_btn").getUnsignedInteger(),
            msg.getField("quick_connector_btn").getUnsignedInteger(),
            msg.getField("start_tars_btn").getUnsignedInteger(),
            msg.getField("arm_switch").getUnsignedInteger());
        return true;
    }
    else if (messageName == "SET_ANTENNA_COORDINATES_ARP_TC")
    {
        mavlink_msg_set_antenna_coordinates_arp_tc_pack(
            sysId, compId, &output, msg.getField("latitude").getDouble(),
            msg.getField("longitude").getDouble(),
            msg.getField("height").getDouble());
        return true;
    }
    else if (messageName == "SET_ROCKET_COORDINATES_ARP_TC")
    {
        mavlink_msg_set_rocket_coordinates_arp_tc_pack(
            sysId, compId, &output, msg.getField("latitude").getDouble(),
            msg.getField("longitude").getDouble(),
            msg.getField("height").getDouble());
        return true;
    }
    else if (messageName == "SET_STEPPER_ANGLE_TC")
    {
        mavlink_msg_set_stepper_angle_tc_pack(
            sysId, compId, &output,
            msg.getField("stepper_id").getUnsignedInteger(),
            msg.getField("angle").getDouble());
        return true;
    }
    else if (messageName == "SET_STEPPER_STEPS_TC")
    {
        mavlink_msg_set_stepper_steps_tc_pack(
            sysId, compId, &output,
            msg.getField("stepper_id").getUnsignedInteger(),
            msg.getField("steps").getDouble());
        return true;
    }
    else if (messageName == "SET_STEPPER_MULTIPLIER_TC")
    {
        mavlink_msg_set_stepper_multiplier_tc_pack(
            sysId, compId, &output,
            msg.getField("stepper_id").getUnsignedInteger(),
            msg.getField("multiplier").getDouble());
        return true;
    }
    else if (messageName == "SET_CALIBRATION_PRESSURE_TC")
    {
        mavlink_msg_set_calibration_pressure_tc_pack(
            sysId, compId, &output, msg.getField("pressure").getDouble());
        return true;
    }
    else if (messageName == "SET_INITIAL_MEA_MASS_TC")
    {
        mavlink_msg_set_calibration_pressure_tc_pack(
            sysId, compId, &output, msg.getField("mass").getDouble());
        return true;
    }

    return false;
}
