/*
 * This file is part of Skyward Hub.
 *
 * Skyward Hub is free software: you can redistribute it and/or modify it under
 * the terms of the GNU General Public License as published by the Free Software
 * Foundation, either version 3 of the License, or (at your option) any later
 * version.
 *
 * Skyward Hub is distributed in the hope that it will be useful, but WITHOUT
 * ANY WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS
 * FOR A PARTICULAR PURPOSE. See the GNU General Public License for more
 * details.
 *
 * You should have received a copy of the GNU General Public License along with
 * Skyward Hub. If not, see <https://www.gnu.org/licenses/>.
 *
 */

#include "ModulesList.h"

#include <Components/FilterSelector/FilterSelector.h>
#include <Components/ModulesPicker/ModulesPicker.h>
#include <Modules/ARPStateViewer/ARPStateViewer.h>
#include <Modules/CommandPad/CommandPad.h>
#include <Modules/CompactCommandPad/CompactCommandPad.h>
#include <Modules/Empty/EmptyModule.h>
#include <Modules/FileStream/FileStreamModule.h>
#include <Modules/Graph/Graph.h>
#include <Modules/IncomingMessagesViewer/IncomingMessagesViewerModule.h>
#include <Modules/MainStateViewer/MainStateViewer.h>
#include <Modules/Mavlink/SerialMavlinkModule.h>
#include <Modules/Mavlink/UdpMavlinkModule.h>
#include <Modules/OrientationVisualizer/OrientationVisualizer.h>
#include <Modules/OutgoingMessagesViewer/OutgoingMessagesViewerModule.h>
#include <Modules/PayloadStateViewer/PayloadStateViewer.h>
#include <Modules/RigStateViewer/RigStateViewer.h>
#include <Modules/Splitter/Splitter.h>
#include <Modules/Tabs/TabsModule.h>
#include <Modules/Test/TestModule.h>
#include <Modules/TimerController/TimerControllerModule.h>
#include <Modules/ValuesConverterViewer/ValuesConverterViewerModule.h>
#include <Modules/ValvesViewer/ValvesViewer.h>

ModulesList& ModulesList::getInstance()
{
    static ModulesList instance;
    return instance;
}

ModulesList::ModulesList()
{

    // If you are creating a new Module, please add it's id in the ModuleId enum
    // (you can find it in moduleslist.h)

    // Hidden
    {
        ModuleInfo emptyModule(ModuleId::EMPTY, "Empty",
                               ModuleCategory::HIDDEN);
        emptyModule.setFactoryMethod([]() { return new EmptyModule(); });
        modulesInfo.append(emptyModule);
    }

    // Data visualization widgets
    {
        ModuleInfo graphModule(ModuleId::GRAPH, "Graph",
                               ModuleCategory::DATA_VISUALIZATION);
        graphModule.setFactoryMethod([]() { return new Graph(); });
        modulesInfo.append(graphModule);

        ModuleInfo outMsgViewer(ModuleId::OUTGOING_MESSAGES_VIEWER,
                                "Outgoing Messages Viewer",
                                ModuleCategory::DATA_VISUALIZATION);
        outMsgViewer.setFactoryMethod(
            []() { return new OutgoingMessagesViewerModule(); });
        modulesInfo.append(outMsgViewer);

        ModuleInfo inMsgViewer(ModuleId::INCOMING_MESSAGES_VIEWER,
                               "Incoming Messages Viewer",
                               ModuleCategory::DATA_VISUALIZATION);
        inMsgViewer.setFactoryMethod(
            []() { return new IncomingMessagesViewerModule(); });
        modulesInfo.append(inMsgViewer);

        ModuleInfo mainStateViewer(ModuleId::MAIN_STATE_VIEWER,
                                   "Main State Viewer",
                                   ModuleCategory::DATA_VISUALIZATION);
        mainStateViewer.setFactoryMethod(
            []() { return new MainStateViewerModule(); });
        modulesInfo.append(mainStateViewer);

        ModuleInfo payloadStateViewer(ModuleId::PAYLOAD_STATE_VIEWER,
                                      "Payload State Viewer",
                                      ModuleCategory::DATA_VISUALIZATION);
        payloadStateViewer.setFactoryMethod(
            []() { return new PayloadStateViewerModule(); });
        modulesInfo.append(payloadStateViewer);

        ModuleInfo rigStateViewer(ModuleId::RIG_STATE_VIEWER,
                                  "Rig State Viewer",
                                  ModuleCategory::DATA_VISUALIZATION);
        rigStateViewer.setFactoryMethod([]()
                                        { return new RigStateViewerModule(); });
        modulesInfo.append(rigStateViewer);

        ModuleInfo ARPStateViewer(ModuleId::ARP_STATE_VIEWER,
                                  "ARP State Viewer",
                                  ModuleCategory::DATA_VISUALIZATION);
        ARPStateViewer.setFactoryMethod([]()
                                        { return new ARPStateViewerModule(); });
        modulesInfo.append(ARPStateViewer);

        ModuleInfo orientationVisualizer(ModuleId::ORIENTATION_VISUALIZER,
                                         "Orientation Visualizer",
                                         ModuleCategory::DATA_VISUALIZATION);
        orientationVisualizer.setFactoryMethod(
            []() { return new OrientationVisualizer(); });
        modulesInfo.append(orientationVisualizer);

        ModuleInfo valvesViewer(ModuleId::VALVES_VIEWER, "Valves Viewer",
                                ModuleCategory::DATA_VISUALIZATION);
        valvesViewer.setFactoryMethod([]() { return new ValvesViewer(); });
        modulesInfo.append(valvesViewer);
    }

    // Data sources
    {
        ModuleInfo serialMavlink(ModuleId::SERIAL_MAVLINK, "Serial Mavlink",
                                 ModuleCategory::DATA_SOURCES);
        serialMavlink.setFactoryMethod([]()
                                       { return new SerialMavlinkModule(); });
        modulesInfo.append(serialMavlink);

        ModuleInfo udpMavlink(ModuleId::UDP_MAVLINK, "Udp Mavlink",
                              ModuleCategory::DATA_SOURCES);
        udpMavlink.setFactoryMethod([]() { return new UdpMavlinkModule(); });
        modulesInfo.append(udpMavlink);

        ModuleInfo fileStream(ModuleId::FILE_STREAM, "File Stream",
                              ModuleCategory::DATA_SOURCES);
        fileStream.setFactoryMethod([]() { return new FileStreamModule(); });
        modulesInfo.append(fileStream);
    }

    // Utilities
    {
        ModuleInfo commandPad(ModuleId::COMMAND_PAD, "Command Pad",
                              ModuleCategory::UTILITIES);
        commandPad.setFactoryMethod([]() { return new CommandPad(); });
        modulesInfo.append(commandPad);

        ModuleInfo compactCommandPad(ModuleId::COMPACT_COMMAND_PAD,
                                     "Compact Command Pad",
                                     ModuleCategory::UTILITIES);
        compactCommandPad.setFactoryMethod([]()
                                           { return new CompactCommandPad(); });
        modulesInfo.append(compactCommandPad);

        ModuleInfo valuesConverterViewer(ModuleId::VALUES_CONVERTER_VIEWER,
                                         "Values Converter Viewer",
                                         ModuleCategory::UTILITIES);
        valuesConverterViewer.setFactoryMethod(
            []() { return new ValuesConverterViewerModule(); });
        modulesInfo.append(valuesConverterViewer);

        ModuleInfo timerController(ModuleId::TIMER_CONTROLLER,
                                   "Timer Controller",
                                   ModuleCategory::UTILITIES);
        timerController.setFactoryMethod(
            []() { return new TimerControllerModule(); });
        modulesInfo.append(timerController);
    }

    // Organization
    {
        ModuleInfo tabs(ModuleId::TABS, "Tabs", ModuleCategory::ORGANIZATION);
        tabs.setFactoryMethod([]() { return new TabsModule(); });
        modulesInfo.append(tabs);

        ModuleInfo verticalSplitter(ModuleId::VERTICAL_SPLITTER,
                                    "Vertical Splitter",
                                    ModuleCategory::ORGANIZATION);
        verticalSplitter.setFactoryMethod(
            []() {
                return new Splitter(Qt::Vertical, ModuleId::VERTICAL_SPLITTER);
            });
        modulesInfo.append(verticalSplitter);

        ModuleInfo horizontalSplitter(ModuleId::HORIZONTAL_SPLITTER,
                                      "Horizontal Splitter",
                                      ModuleCategory::ORGANIZATION);
        horizontalSplitter.setFactoryMethod(
            []() {
                return new Splitter(Qt::Horizontal,
                                    ModuleId::HORIZONTAL_SPLITTER);
            });
        modulesInfo.append(horizontalSplitter);
    }

    // Testing
    {
        ModuleInfo testModule(ModuleId::BROKER_TEST, "Test Broker",
                              ModuleCategory::TESTING);
        testModule.setFactoryMethod([]() { return new TestModule(); });
        modulesInfo.append(testModule);
    }
}

QString ModulesList::getModuleName(ModuleId id)
{
    for (int i = 0; i < modulesInfo.count(); i++)
    {
        if (modulesInfo[i].getId() == id)
            return modulesInfo[i].getModuleName();
    }

    return "";
}

bool ModulesList::getModuleId(QString name, ModuleId& id)
{
    for (int i = 0; i < modulesInfo.count(); i++)
    {
        if (modulesInfo[i].getModuleName().toLower().replace(' ', '_') ==
            name.toLower().replace(' ', '_'))
        {
            id = modulesInfo[i].getId();
            return true;
        }
    }

    return false;
}

QString ModulesList::getXmlName(ModuleId id)
{
    QString name = getModuleName(id);
    name         = name.toLower().replace(' ', '_');
    return name;
}

std::function<Module*()> ModulesList::getFactoryMethod(ModuleId id)
{
    for (int i = 0; i < modulesInfo.count(); i++)
    {
        if (modulesInfo[i].getId() == id)
            return modulesInfo[i].getFactoryMethod();
    }
    return []() { return nullptr; };
}

std::function<Module*()> ModulesList::getFactoryMethod(const QString& name)
{
    for (int i = 0; i < modulesInfo.count(); i++)
    {
        if (modulesInfo[i].getModuleName().toLower().replace(' ', '_') ==
            name.toLower().replace(' ', '_'))
            return modulesInfo[i].getFactoryMethod();
    }
    return []() { return nullptr; };
}

QList<QString> ModulesList::getModulesNamesList()
{
    QList<QString> names;
    for (int i = 0; i < modulesInfo.count(); i++)
    {
        names.append(modulesInfo[i].getModuleName());
    }
    return names;
}

QList<ModuleInfo> ModulesList::getModulesInfo() { return modulesInfo; }

Module* ModulesList::instantiateModule(ModuleId id)
{
    return getFactoryMethod(id)();
}

Module* ModulesList::instantiateModule(const QString& name)
{
    return getFactoryMethod(name)();
}
