/*
 * This file is part of Skyward Hub.
 *
 * Skyward Hub is free software: you can redistribute it and/or modify it under
 * the terms of the GNU General Public License as published by the Free Software
 * Foundation, either version 3 of the License, or (at your option) any later
 * version.
 *
 * Skyward Hub is distributed in the hope that it will be useful, but WITHOUT
 * ANY WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS
 * FOR A PARTICULAR PURPOSE. See the GNU General Public License for more
 * details.
 *
 * You should have received a copy of the GNU General Public License along with
 * Skyward Hub. If not, see <https://www.gnu.org/licenses/>.
 *
 */

#include "TabsModule.h"

#include <Modules/ModulesList.h>

#include <QBoxLayout>
#include <QDebug>
#include <QLabel>

TabsModule::TabsModule() : Module(ModuleId::TABS)
{
    setupUi();
    customContextMenuActionSetup();

    QObject::connect(tabNames, &QListWidget::currentRowChanged, tabContents,
                     &QStackedWidget::setCurrentIndex);
    QObject::connect(tabNames, &QListWidget::itemDoubleClicked, this,
                     &TabsModule::onDoubleClickedItem);

    // Automatically add first tab
    addTab(generateUniqueTabName(),
           ModulesList::getInstance().instantiateModule(ModuleId::EMPTY));
}

XmlObject TabsModule::toXmlObject()
{
    XmlObject obj = Module::toXmlObject();

    for (int i = 0; i != tabContents->count(); i++)
    {
        XmlObject tab("tab");
        tab.addAttribute("name", tabNames->item(i)->text());
        if (auto module = qobject_cast<Module*>(tabContents->widget(i)))
        {
            tab.addChild(module->toXmlObject());
        }
        else
        {
            qCritical() << "Unable to save tab " << tabNames->item(i)->text()
                        << " due to failed when casting widget to Module*";
            continue;  // Skip this tab
        }

        obj.addChild(tab);
    }

    return obj;
}

void TabsModule::fromXmlObject(const XmlObject& xmlObject)
{
    // Discard previous content
    tabNames->clear();
    for (int i = tabContents->count() - 1; i >= 0; i--)
        delete tabContents->widget(i);

    for (int i = 0; i < xmlObject.childCount(); ++i)
    {
        XmlObject tabNode = xmlObject.childAt(i);
        if (tabNode.getObjectName() != "tab" || tabNode.childCount() == 0)
            continue;

        XmlObject xmlChild = tabNode.childAt(0);

        Module* module = ModulesList::getInstance().instantiateModule(
            xmlChild.getObjectName());
        if (module != nullptr)
        {
            module->fromXmlObject(xmlChild);
            addTab(tabNode.getAttribute("name", generateUniqueTabName()),
                   module);
        }
        else
        {
            qCritical() << "Failed to instantiate module"
                        << tabNode.getAttribute("name",
                                                generateUniqueTabName());
        }
    }

    // If the tabs are left empty, add a single tab with an empty module
    if (tabNames->count() == 0)
        addTab(generateUniqueTabName(),
               ModulesList::getInstance().instantiateModule(ModuleId::EMPTY));

    tabNames->setCurrentItem(tabNames->item(0));
}

void TabsModule::addTab(const QString& tabName, Module* module)
{
    connect(module, &Module::replaceMe, this, &TabsModule::replaceTab);
    connect(module, &Module::closeMe, this, &TabsModule::closeTab);

    auto* listItem = new QListWidgetItem();
    listItem->setText(tabName);
    listItem->setTextAlignment(Qt::AlignCenter);
    tabNames->addItem(listItem);

    tabContents->addWidget(module);

    // We set the correct tabName after the widget is added
    tabNames->setCurrentItem(listItem);
}

QString TabsModule::generateUniqueTabName()
{
    return QString("Tab %1").arg(tabNames->count());
}

void TabsModule::replaceTab(Module* oldModule, Module* newModule)
{
    // Check if the modules are valid
    if (!oldModule || !newModule)
        return;

    // Check if we are trying to replace a module with itself
    if (oldModule == newModule)
        return;

    // Get the index of the old module
    int index = tabContents->indexOf(oldModule);
    if (index == -1)
        return;

    // Replace the module
    tabContents->insertWidget(index, newModule);
    tabContents->setCurrentWidget(newModule);
    tabContents->removeWidget(oldModule);
    delete oldModule;
    tabContents->update();

    connect(newModule, &Module::replaceMe, this, &TabsModule::replaceTab);
    connect(newModule, &Module::closeMe, this, &TabsModule::closeTab);
}

void TabsModule::closeTab(Module* module)
{
    int index = tabContents->indexOf(module);

    // Check if the given module is a child
    if (index == -1)
        return;

    // If there is only one child, then close the entire tabs module
    if (tabContents->count() == 1)
    {
        emit closeMe(this);
        return;
    }

    tabContents->removeWidget(module);
    delete tabNames->item(index);
    tabNames->setCurrentItem(tabNames->item(index == 0 ? 0 : index - 1));
}

void TabsModule::onDoubleClickedItem(QListWidgetItem* item)
{
    item->setFlags(item->flags() | Qt::ItemIsEditable);
    tabNames->setCurrentItem(item);
    tabNames->editItem(item);
}

void TabsModule::customContextMenuActionSetup()
{
    auto* newTab = new QAction("New Tab");
    connect(newTab, &QAction::triggered, this, &TabsModule::onMenuNewTabClick);
    customContextMenuActions.append(newTab);

    auto* deleteTab = new QAction("Delete Tab");
    connect(deleteTab, &QAction::triggered, this,
            &TabsModule::onMenuDeleteTabClick);
    customContextMenuActions.append(deleteTab);

    auto* renameTab = new QAction("Rename Tab");
    connect(renameTab, &QAction::triggered, this,
            &TabsModule::onMenuRenameTabClick);
    customContextMenuActions.append(renameTab);
}

void TabsModule::onMenuNewTabClick()
{
    addTab(generateUniqueTabName(),
           ModulesList::getInstance().instantiateModule(ModuleId::EMPTY));
}

void TabsModule::onMenuDeleteTabClick()
{
    if (auto module = qobject_cast<Module*>(tabContents->currentWidget()))
        closeTab(module);
    else
        qCritical() << "Unable to close tab  due to failed when casting widget "
                       "to Module*";
}

void TabsModule::onMenuRenameTabClick()
{
    auto* item = tabNames->currentItem();
    item->setFlags(item->flags() | Qt::ItemIsEditable);
    tabNames->editItem(item);
}

void TabsModule::setupUi()
{
    // Top component (top bar)
    tabNames = new QListWidget;
    tabNames->setFlow(QListView::Flow::LeftToRight);
    tabNames->setSizePolicy(QSizePolicy::Expanding, QSizePolicy::Preferred);
    tabNames->setSizeAdjustPolicy(QAbstractScrollArea::AdjustToContents);
    tabNames->setSelectionMode(QAbstractItemView::SingleSelection);
    tabNames->setMaximumHeight(40);

    // Bottom component
    tabContents = new QStackedWidget;
    tabContents->setObjectName("TabsStackedWidget");

    // Putting all together
    QBoxLayout* outerLayout = new QBoxLayout(QBoxLayout::TopToBottom, this);
    outerLayout->setContentsMargins(0, 0, 0, 0);
    outerLayout->setSpacing(0);
    outerLayout->addWidget(tabNames);
    outerLayout->addWidget(tabContents);
}
